<?php

namespace Drupal\commerce_avatax\EventSubscriber;

use Drupal\commerce_order\Adjustment;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\state_machine\Event\WorkflowTransitionEvent;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

class CommitTransactionSubscriber implements EventSubscriberInterface {

  /**
   * The tax type storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $taxTypeStorage;

  /**
   * Constructs a new CommitTransactionSubscriber object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager) {
    $this->taxTypeStorage = $entity_type_manager->getStorage('commerce_tax_type');
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    $events = [
      'commerce_order.place.post_transition' => ['commitTransaction'],
    ];
    return $events;
  }

  /**
   * Commits a transaction or the order in AvaTax.
   *
   * @param \Drupal\state_machine\Event\WorkflowTransitionEvent $event
   *   The workflow transition event.
   */
  public function commitTransaction(WorkflowTransitionEvent $event) {
    /** @var \Drupal\commerce_order\Entity\OrderInterface $order */
    $order = $event->getEntity();
    /** @var \Drupal\commerce_order\Adjustment[] $avatax_adjustments */
    $avatax_adjustments = array_filter($order->collectAdjustments(), function (Adjustment $adjustment) use ($order) {
      return $adjustment->getType() == 'tax'
        && strpos($adjustment->getSourceId(), 'avatax|') !== FALSE
        && $order->getData($adjustment->getSourceId());
    });

    foreach ($avatax_adjustments as $avatax_adjustment) {
      list(, $tax_type_id) = explode('|', $avatax_adjustment->getSourceId(), 2);
      /** @var \Drupal\commerce_tax\Entity\TaxTypeInterface $tax_type */
      $tax_type = $this->taxTypeStorage->load($tax_type_id);
      /** @var \Drupal\commerce_avatax\Plugin\Commerce\TaxType\Avatax $tax_type_plugin */
      $tax_type_plugin = $tax_type->getPlugin();

      if ($tax_type_plugin->getConfiguration()['disable_commit']) {
        continue;
      }

      $avatax_transaction = $order->getData($avatax_adjustment->getSourceId());
      $request_body = $avatax_transaction['request'];
      $request_body['type'] = 'SalesInvoice';
      $request_body['commit'] = TRUE;

      $tax_type_plugin->getClient()->post('/api/v2/transactions/create', [
        'json' => $request_body,
      ]);
    }
  }

}
