<?php

namespace Drupal\commerce_avatax\Plugin\Commerce\TaxType;

use Drupal\address\AddressInterface;
use Drupal\commerce_avatax\ClientFactory;
use Drupal\commerce_avatax\Resolver\ChainTaxCodeResolverInterface;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_order\Adjustment;
use Drupal\commerce_price\Price;
use Drupal\commerce_tax\Plugin\Commerce\TaxType\RemoteTaxTypeBase;
use Drupal\Component\Serialization\Json;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use GuzzleHttp\Exception\ClientException;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;

/**
 * Provides the Avatax remote tax type.
 *
 * @CommerceTaxType(
 *   id = "avatax",
 *   label = "Avatax",
 * )
 */
class Avatax extends RemoteTaxTypeBase {

  /**
   * The client.
   *
   * @var \GuzzleHttp\Client
   */
  protected $client;

  /**
   * The chain tax code resolver.
   *
   * @var \Drupal\commerce_avatax\Resolver\ChainTaxCodeResolverInterface
   */
  protected $chainTaxCodeResolver;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The logger.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * Constructs a new Avatax object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Symfony\Component\EventDispatcher\EventDispatcherInterface $event_dispatcher
   *   The event dispatcher.
   * @param \Drupal\commerce_avatax\ClientFactory $client_factory
   *   The client.
   * @param \Drupal\commerce_avatax\Resolver\ChainTaxCodeResolverInterface $chain_tax_code_resolver
   *   The chain tax code resolver.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entity_type_manager, EventDispatcherInterface $event_dispatcher, ClientFactory $client_factory, ChainTaxCodeResolverInterface $chain_tax_code_resolver, ModuleHandlerInterface $module_handler, LoggerChannelFactoryInterface $logger_factory) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $entity_type_manager, $event_dispatcher);
    $this->client = $client_factory->createInstance($this->configuration);
    $this->chainTaxCodeResolver = $chain_tax_code_resolver;
    $this->moduleHandler = $module_handler;
    $this->logger = $logger_factory->get('commerce_avatax');
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('event_dispatcher'),
      $container->get('commerce_avatax.client_factory'),
      $container->get('commerce_avatax.chain_tax_code_resolver'),
      $container->get('module_handler'),
      $container->get('logger.factory')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'display_inclusive' => FALSE,
      'account_id' => '',
      'license_key' => '',
      'api_mode' => 'development',
      'company_code' => 'DEFAULT',
      'disable_commit' => FALSE,
      'shipping_tax_code' => 'FR020100',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $id_prefix = implode('-', $form['#parents']);
    // @todo We cannot use unique IDs, or multiple elements on a page currently.
    // @see https://www.drupal.org/node/2675688
    // $wrapper_id = Html::getUniqueId($id_prefix . '-ajax-wrapper');
    $wrapper_id = $id_prefix . '-ajax-wrapper';
    $form['configuration'] = [
      '#type' => 'details',
      '#title' => $this->t('Configuration'),
      '#open' => TRUE,
      '#prefix' => '<div id="' . $wrapper_id . '">',
      '#suffix' => '</div>',
      // Pass the id along to other methods.
      '#wrapper_id' => $wrapper_id,
    ];
    $form['configuration']['api_mode'] = [
      '#type' => 'radios',
      '#title' => $this->t('API mode:'),
      '#default_value' => $this->configuration['api_mode'],
      '#options' => [
        'development' => $this->t('Development'),
        'production' => $this->t('Production'),
      ],
      '#required' => TRUE,
      '#description' => $this->t('The mode to use when calculating taxes.'),
    ];
    $form['configuration']['account_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Account ID:'),
      '#default_value' => $this->configuration['account_id'],
      '#required' => TRUE,
      '#description' => $this->t('The account ID to use when calculating taxes.'),
    ];
    $form['configuration']['license_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('License key:'),
      '#default_value' => $this->configuration['license_key'],
      '#required' => TRUE,
      '#description' => $this->t('The license key to send to Avatax when calculating taxes.'),
    ];
    $form['configuration']['company_code'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Company code:'),
      '#default_value' => $this->configuration['company_code'],
      '#required' => TRUE,
      '#description' => $this->t('The default company code to send to Avatax when calculating taxes. Using <em>DEFAULT</em> will use the default store configured in Avatax.'),
    ];

    $form['configuration']['validate'] = [
      '#type' => 'submit',
      '#value' => t('Validate credentials'),
      '#limit_validation_errors' => [],
      '#ajax' => [
        'callback' => [$this, 'validateCredentials'],
        'wrapper' => $wrapper_id,
      ],
    ];

    $form['configuration']['disable_commit'] = [
      '#type' => 'checkbox',
      '#title' => t('Disable document committing.'),
      '#description' => t('Enable this option if you are only using the AvaTax service to display taxes and a backend system is performing the final commit of the tax document.'),
      '#default_value' => $this->configuration['disable_commit'],
    ];

    $form['configuration']['shipping_tax_code'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Shipping tax code:'),
      '#default_value' => $this->configuration['shipping_tax_code'],
      '#required' => TRUE,
      '#description' => $this->t('The taxCode to use for each shipment line item.'),
      '#access' => $this->moduleHandler->moduleExists('commerce_shipping'),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::validateConfigurationForm($form, $form_state);
    $values = $form_state->getValue($form['#parents']);

    try {
      $ping_request = $this->client->get('/api/v2/utilities/ping', [
        'headers' => [
          'Authorization' => 'Basic ' . base64_encode($values['configuration']['account_id'] . ':' . $values['configuration']['license_key']),
        ],
      ]);
      $ping_request = Json::decode($ping_request->getBody()->getContents());
      if (!empty($ping_request['authenticated']) && $ping_request['authenticated'] === TRUE) {
        drupal_set_message(t('AvaTax response confirmed using the account and license key above.'));
      }
      else {
        $form_state->setErrorByName('configuration][avatax][configuration][account_id', t('Could not confirm the provided credentials.'));
        $form_state->setErrorByName('configuration][avatax][configuration][license_key');
      }
    }
    catch (\Exception $e) {
      $form_state->setErrorByName('configuration][avatax][configuration', $e->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);
    if (!$form_state->getErrors()) {
      $values = $form_state->getValue($form['#parents']);
      $this->configuration['api_mode'] = $values['configuration']['api_mode'];
      $this->configuration['account_id'] = $values['configuration']['account_id'];
      $this->configuration['license_key'] = $values['configuration']['license_key'];
      $this->configuration['company_code'] = $values['configuration']['company_code'];
    }
  }

  /**
   * Ajax callback for validation.
   */
  public function validateCredentials(array &$form, FormStateInterface $form_state) {
    $triggering_element = $form_state->getTriggeringElement();
    return NestedArray::getValue($form, array_slice($triggering_element['#array_parents'], 0, -1));
  }

  /**
   * {@inheritdoc}
   */
  public function apply(OrderInterface $order) {
    // Attempt to get company code for specific store.
    $store = $order->getStore();
    if ($store->get('avatax_company_code')->isEmpty()) {
      $company_code = $this->configuration['company_code'];
    }
    else {
      $company_code = $store->get('avatax_company_code')->value;
    }

    $currency_code = $order->getTotalPrice() ? $order->getTotalPrice()->getCurrencyCode() : $store->getDefaultCurrencyCode();

    $request_body = [
      'type' => 'SalesOrder',
      'companyCode' => $company_code,
      'date' => $this->getCalculationDate($order)->format('c'),
      'code' => 'DC-' . $order->id(),
      'customerCode' => $order->getEmail(),
      'currencyCode' => $currency_code,
      'lines' => [],
    ];
    $has_shipments = $order->hasField('shipments') && !$order->get('shipments')->isEmpty();

    foreach ($order->getItems() as $item) {
      $profile = $this->resolveCustomerProfile($item);

      // If we could not resolve a profile for the order item, do not add it
      // to the API request. There may not be an address available yet, or the
      // item may not be shippable and not attached to a shipment.
      if (!$profile) {
        continue;
      }

      /** @var \Drupal\address\Plugin\Field\FieldType\AddressItem $address */
      $address = $profile->get('address')->first();
      $line_item = [
        'number' => $item->uuid(),
        'quantity' => $item->getQuantity(),
        'amount' => $item->getAdjustedTotalPrice()->getNumber(),
      ];
      if ($has_shipments) {
        $line_item['addresses'] = [
          'shipFrom' => self::formatAddress($store->getAddress()),
          'shipTo' => self::formatAddress($address),
        ];
      }
      else {
        $line_item['addresses']['singleLocation'] = self::formatAddress($address);
      }

      $line_item['taxCode'] = $this->chainTaxCodeResolver->resolve($item);
      $request_body['lines'][] = $line_item;
    }

    if ($has_shipments) {
      /** @var \Drupal\commerce_shipping\Entity\ShipmentInterface $shipment */
      foreach ($order->get('shipments')->referencedEntities() as $shipment) {
        $request_body['lines'][] = [
          'taxCode' => $this->configuration['shipping_tax_code'],
          'number' => $shipment->uuid(),
          'description' => $shipment->label(),
          'amount' => $shipment->getAmount()->getNumber(),
          'quantity' => 1,
          'addresses' => [
            'shipFrom' => self::formatAddress($store->getAddress()),
            'shipTo' => self::formatAddress($shipment->getShippingProfile()->get('address')->first()),
          ]
        ];
      }
    }

    $this->moduleHandler->alter('commerce_avatax_order_request', $request_body, $order);

    // Do not go further unless there have been lines added.
    if (empty($request_body['lines'])) {
      return;
    }

    try {
      $response = $this->client->post('/api/v2/transactions/create', [
        'json' => $request_body,
      ]);

      $response_body = Json::decode($response->getBody()->getContents());

      $adjustments = [];
      $applied_adjustments = [];
      foreach ($response_body['lines'] as $tax_adjustment) {
        $adjustments[$tax_adjustment['lineNumber']] = $tax_adjustment['tax'];
      }

      // Add tax adjustments to order items.
      foreach ($order->getItems() as $item) {
        if (isset($adjustments[$item->uuid()])) {
          $item->addAdjustment(new Adjustment([
            'type' => 'tax',
            'label' => $this->t('Sales tax'),
            'amount' => new Price((string) $adjustments[$item->uuid()], $currency_code),
            'source_id' => $this->pluginId . '|' . $this->entityId,
          ]));
          $applied_adjustments[$item->uuid()] = $item->uuid();
        }
      }

      // If we still have Tax adjustments to apply, apply one to the order.
      $remaining_adjustments = array_diff_key($adjustments, $applied_adjustments);
      if ($remaining_adjustments) {
        $order->addAdjustment(new Adjustment([
          'type' => 'tax',
          'label' => $this->t('Sales tax'),
          'amount' => new Price((string) array_sum($remaining_adjustments), $currency_code),
          'source_id' => $this->pluginId . '|' . $this->entityId,
        ]));
      }

      // Let other modules respond to the response.
      $this->moduleHandler->alter('commerce_avatax_order_response', $response_body, $order);

      // Store the Avalara data in the order.
      $order->setData($this->pluginId . '|' . $this->entityId, [
        'request' => $request_body,
        'response' => $response_body,
        'adjustments' => $adjustments,
      ]);
    }
    catch (ClientException $e) {
      // @todo port error handling from D7.
      $this->logger->error($e->getResponse()->getBody()->getContents());
    }
    catch (\Exception $e) {
      $this->logger->error($e->getMessage());
    }
  }

  /**
   * Format an address for use in the order request.
   *
   * @param \Drupal\address\AddressInterface $address
   *   The address to format.
   *
   * @return array
   *   Return a formatted address for use in the order request.
   */
  protected static function formatAddress(AddressInterface $address) {
    return [
      'line1' => $address->getAddressLine1(),
      'line2' => $address->getAddressLine2(),
      'city' => $address->getLocality(),
      'region' => $address->getAdministrativeArea(),
      'country' => $address->getCountryCode(),
      'postalCode' => $address->getPostalCode(),
    ];
  }

  /**
   * Get the API client, configured for this instance.
   *
   * @return \GuzzleHttp\Client
   */
  public function getClient() {
    return $this->client;
  }

}
