<?php

namespace Drupal\Tests\commerce_avatax\Kernel;

use Drupal\commerce\CommerceSinglePluginCollection;
use Drupal\commerce_avatax\ClientFactory;
use Drupal\commerce_avatax\Plugin\Commerce\TaxType\Avatax;
use Drupal\commerce_order\Entity\Order;
use Drupal\commerce_order\Entity\OrderItem;
use Drupal\commerce_price\Price;
use Drupal\commerce_product\Entity\Product;
use Drupal\commerce_product\Entity\ProductVariation;
use Drupal\commerce_product\Entity\ProductVariationType;
use Drupal\commerce_tax\Entity\TaxType;
use Drupal\Component\Serialization\Json;
use Drupal\Core\DependencyInjection\ContainerBuilder;
use Drupal\Core\DependencyInjection\ServiceModifierInterface;
use Drupal\Core\Entity\EntityStorageException;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\profile\Entity\Profile;
use Drupal\Tests\commerce\Kernel\CommerceKernelTestBase;
use GuzzleHttp\Client;
use GuzzleHttp\Handler\MockHandler;
use GuzzleHttp\HandlerStack;
use GuzzleHttp\Psr7\Response;
use Prophecy\Argument;
use Psr\Http\Message\RequestInterface;
use Psr\Log\LoggerInterface;

/**
 * Tests the tax type plugin.
 *
 * @group commerce_avatax
 */
class TaxTypePluginTest extends CommerceKernelTestBase implements ServiceModifierInterface {

  /**
   * The tax type plugin.
   *
   * @var \Drupal\commerce_tax\Plugin\Commerce\TaxType\TaxTypeInterface
   */
  protected $plugin;

  /**
   * {@inheritdoc}
   */
  public static $modules = [
    'entity_reference_revisions',
    'path',
    'profile',
    'state_machine',
    'commerce_product',
    'commerce_order',
    'commerce_tax',
    'commerce_avatax',
    'commerce_avatax_test',
  ];

  /**
   * The config.
   *
   * @var array
   */
  protected $config;

  /**
   * A sample order.
   *
   * @var \Drupal\commerce_order\Entity\OrderInterface
   */
  protected $order;

  /**
   * The test tax type.
   *
   * @var \Drupal\commerce_tax\Entity\TaxTypeInterface
   */
  protected $taxType;

  /**
   * {@inheritdoc}
   */
  protected function setUp() {
    parent::setUp();

    $this->installEntitySchema('profile');
    $this->installEntitySchema('commerce_order');
    $this->installEntitySchema('commerce_order_item');
    $this->installEntitySchema('commerce_product');
    $this->installEntitySchema('commerce_product_variation');
    $this->installConfig(['commerce_product', 'commerce_order']);

    $user = $this->createUser(['mail' => $this->randomString() . '@example.com']);

    // Turn off title generation to allow explicit values to be used.
    $variation_type = ProductVariationType::load('default');
    $variation_type->setGenerateTitle(FALSE);
    $variation_type->save();

    $product = Product::create([
      'type' => 'default',
      'title' => 'Default testing product',
    ]);
    $product->save();

    $variation1 = ProductVariation::create([
      'type' => 'default',
      'sku' => 'TEST_' . strtolower($this->randomMachineName()),
      'title' => $this->randomString(),
      'status' => 1,
      'price' => new Price('12.00', 'USD'),
    ]);
    $variation1->save();
    $product->addVariation($variation1)->save();

    $profile = Profile::create([
      'type' => 'customer',
      'address' => [
        'country_code' => 'US',
        'postal_code' => '53177',
        'locality' => 'Milwaukee',
        'address_line1' => 'Pabst Blue Ribbon Dr',
        'administrative_area' => 'WI',
        'given_name' => 'Frederick',
        'family_name' => 'Pabst',
      ],
      'uid' => $user->id(),
    ]);
    $profile->save();
    $profile = $this->reloadEntity($profile);

    /** @var \Drupal\commerce_order\Entity\Order $order */
    $order = Order::create([
      'type' => 'default',
      'state' => 'draft',
      'mail' => $user->getEmail(),
      'uid' => $user->id(),
      'ip_address' => '127.0.0.1',
      'order_number' => '6',
      'billing_profile' => $profile,
      'store_id' => $this->store->id(),
    ]);
    /** @var \Drupal\commerce_order\Entity\OrderItemInterface $order_item */
    $order_item = OrderItem::create([
      'type' => 'default',
      'quantity' => 1,
      'unit_price' => new Price('12.00', 'USD'),
      'purchased_entity' => $variation1,
    ]);
    $order_item->save();
    $order->addItem($order_item);
    $order->save();
    $this->order = $this->reloadEntity($order);

    $this->config = [
      'account_id' => 'DUMMY ACCOUNT',
      'license_key' => 'DUMMY KEY',
      'api_mode' => 'development',
    ];

    $tax_type = TaxType::create([
      'id' => 'avatax_test',
      'label' => 'AvaTax Test',
      'plugin' => 'avatax',
      'configuration' => $this->config
    ]);
    $tax_type->save();
    $this->taxType = $this->reloadEntity($tax_type);
  }

  /**
   * Tests the default configuration assertions.
   *
   * Need to ensure DEFAULT is the company code.
   */
  public function testPluginConfiguration() {
    $this->createPluginMock();
    $this->assertEquals([
      'account_id' => '',
      'license_key' => '',
      'api_mode' => 'development',
      'company_code' => 'DEFAULT',
      'display_inclusive' => FALSE,
      'disable_commit' => FALSE,
      'shipping_tax_code' => 'FR020100',
    ], $this->plugin->defaultConfiguration());
    $this->assertEquals([
      'account_id' => 'DUMMY ACCOUNT',
      'license_key' => 'DUMMY KEY',
      'api_mode' => 'development',
      'company_code' => 'DEFAULT',
      'display_inclusive' => FALSE,
      'disable_commit' => FALSE,
      'shipping_tax_code' => 'FR020100',
    ], $this->plugin->getConfiguration());
  }

  /**
   * Tests that Avalara is not invoked when the billing profile is empty.
   */
  public function testNoRequestWhenBillingEmpty() {
    // Make sure no requests are queued, so an exception will be thrown if
    // a request is created.
    $this->createPluginMock([]);

    $this->order->get('billing_profile')->setValue(NULL);
    $this->assertTrue($this->plugin->applies($this->order));
    $this->plugin->apply($this->order);
    $this->order = $this->reloadEntity($this->order);
    $adjustments = $this->order->collectAdjustments();
    $this->assertCount(0, $adjustments);
  }

  /**
   * Tests applying.
   */
  public function testApply() {
    $this->createPluginMock([
      new Response(200, [], Json::encode([
        'lines' => [
          [
            'lineNumber' => 1,
            'tax' => 5.25,
          ]
        ],
      ]))
    ]);
    $this->assertTrue($this->plugin->applies($this->order));
    $this->plugin->apply($this->order);
    $this->assertNotEmpty($this->order->getData('avatax|avatax_test'));
    $adjustments = $this->order->collectAdjustments();
    $this->assertCount(1, $adjustments);
    $adjustment = reset($adjustments);
    $this->assertEquals('tax', $adjustment->getType());
    $this->assertEquals('Sales tax', $adjustment->getLabel());
    $this->assertEquals('avatax|avatax_test', $adjustment->getSourceId());
    $this->assertEquals(new Price('5.25', 'USD'), $adjustment->getAmount());
  }

  /**
   * Tests that the transaction type is correct when applying the adjustment.
   *
   * @link https://developer.avalara.com/avatax/dev-guide/transactions/should-i-commit/
   */
  public function testTransactionType() {
    $this->createPluginMock([
      new Response(200, [], Json::encode([
        'lines' => [
          [
            'lineNumber' => 1,
            'tax' => 5.25,
          ]
        ],
      ])),
    ], [
      function (callable $handler) {
        return function (RequestInterface $request, array $options) use ($handler) {
          $body = $request->getBody()->getContents();
          $body = Json::decode($body);
          if (empty($body['type']) || $body['type'] == 'SalesInvoice') {
            throw new \Exception('Transaction type should be SalesOrder unless committing the transaction.');
          }
          return $handler($request, $options);
        };
      }
    ]);
    $this->plugin->apply($this->order);
    $this->assertNotEmpty($this->order->getData('avatax|avatax_test'));
  }

  /**
   * Tests that a transaction is committed when an order is placed.
   */
  public function testCommitTransaction() {
    $this->createPluginMock([
      new Response(200, [], Json::encode([
        'lines' => [
          [
            'lineNumber' => 1,
            'tax' => 5.25,
          ]
        ],
      ])),
      new Response(200, [], Json::encode([
        'lines' => [
          [
            'lineNumber' => 1,
            'tax' => 5.25,
          ]
        ],
      ]))
    ], [
      function (callable $handler) {
        return function (RequestInterface $request, array $options) use ($handler) {
          $count = \Drupal::state()->get('avatax_request_count', 0);
          $count++;
          \Drupal::state()->set('avatax_request_count', $count);
          if ($count == 2) {
            $body = $request->getBody()->getContents();
            $body = Json::decode($body);
            if (empty($body['type']) || $body['type'] != 'SalesInvoice') {
              throw new \Exception('Transaction type should be SalesInvoice when committing the transaction.');
            }
            if (empty($body['commit']) || !$body['commit']) {
              throw new \Exception('Transaction type be committed if SalesInvoice.');
            }
          }
          return $handler($request, $options);
        };
      }
    ]);
    $this->plugin->apply($this->order);
    $transition = $this->order->getState()->getTransitions()['place'];
    $this->order->getState()->applyTransition($transition);

    $this->order->save();

    $this->assertEquals(2, $this->container->get('state')->get('avatax_request_count'));
  }

  /**
   * Tests that a transaction is not committed when configured to skip.
   */
  public function testDisableCommitTransaction() {
    $this->config['disable_commit'] = TRUE;
    $this->createPluginMock([
      new Response(200, [], Json::encode([
        'lines' => [
          [
            'lineNumber' => 1,
            'tax' => 5.25,
          ]
        ],
      ]))
    ]);
    $this->plugin->apply($this->order);
    $transition = $this->order->getState()->getTransitions()['place'];
    $this->order->getState()->applyTransition($transition);

    // Will not error due to empty queue.
    $this->order->save();
  }

  /**
   * Tests the case code resolver.
   */
  public function testTaxCodeResolver() {
    $this->createPluginMock([
      new Response(200, [], Json::encode([
        'lines' => [
          [
            'lineNumber' => 1,
            'tax' => 5.25,
          ]
        ],
      ])),
      new Response(200, [], Json::encode([
        'lines' => [
          [
            'lineNumber' => 1,
            'tax' => 5.25,
          ]
        ],
      ])),
    ]);
    $this->plugin->apply($this->order);
    $request_data = $this->order->getData('avatax|avatax_test');

    $this->assertNull($request_data['request']['lines'][0]['taxCode']);


    $variation2 = ProductVariation::create([
      'type' => 'default',
      'sku' => 'TEST_' . strtolower($this->randomMachineName()),
      'title' => $this->randomString(),
      'status' => 1,
      'price' => new Price('12.00', 'USD'),
      'avatax_tax_code' => 'TESTCODE123',
    ]);
    $variation2->save();
    $order_item = OrderItem::create([
      'type' => 'default',
      'quantity' => 1,
      'unit_price' => new Price('12.00', 'USD'),
      'purchased_entity' => $variation2,
    ]);
    $order_item->save();
    $this->order->addItem($order_item);
    $this->order->save();

    $request_data = $this->order->getData('avatax|avatax_test');

    $this->assertNull($request_data['request']['lines'][0]['taxCode']);
    $this->assertEquals('TESTCODE123', $request_data['request']['lines'][1]['taxCode']);
  }

  /**
   * {@inheritdoc}
   */
  public function alter(ContainerBuilder $container) {
    $container->removeDefinition('test.http_client.middleware');
  }

  /**
   * Creates a plugin mock.
   *
   * This plugin mock will throw exceptions when an error is logged and allows
   * for mocking responses.
   *
   * @param \Psr\Http\Message\ResponseInterface[] $responses
   *   An array of mocked responses.
   * @param callable[] $middlewares
   *   An array of callable handlers.
   *
   * @throws \Exception
   */
  protected function createPluginMock(array $responses = [], array $middlewares = []) {
    $mock_logger = $this->prophesize(LoggerInterface::class);
    $mock_logger->error(Argument::any())->will(function ($args) {
      throw new \Exception($args[0]);
    });
    $mock_logger_factory = $this->prophesize(LoggerChannelFactoryInterface::class);
    $mock_logger_factory->get('commerce_avatax')->willReturn($mock_logger->reveal());


    $mock_handler = new MockHandler($responses);
    $mock_handler_stack = HandlerStack::create($mock_handler);
    foreach ($middlewares as $middleware) {
      $mock_handler_stack->push($middleware);
    }
    $mock_client = new Client([
      'handler' => $mock_handler_stack,
      'base_uri' => 'https://sandbox-rest.avatax.com/',
      'headers' => [
        'Authorization' => 'Basic ' . base64_encode($this->config['account_id'] . ':' . $this->config['license_key']),
        'Content-Type' => 'application/json',
        'x-Avalara-UID' => 'a0o33000003waOC',
        'x-Avalara-Client' => 'Test Client',
      ],
    ]);
    $mock_client_factory = $this->prophesize(ClientFactory::class);
    $mock_client_factory->createInstance(Argument::any())->willReturn($mock_client);

    $this->plugin = new Avatax(
      $this->config + ['_entity_id' => 'avatax_test'],
      'avatax',
      ['label' => 'Avatax'],
      $this->container->get('entity_type.manager'),
      $this->container->get('event_dispatcher'),
      $mock_client_factory->reveal(),
      $this->container->get('commerce_avatax.chain_tax_code_resolver'),
      $this->container->get('module_handler'),
      $mock_logger_factory->reveal()
    );

    // @todo Figure out how to support the mocked plugin.
    // This will not work, as config entities are not statically cached. So
    // any later ::load will refresh the object.
    $plugin_collection = new CommerceSinglePluginCollection(
      $this->container->get('plugin.manager.commerce_tax_type'),
      'avatax',
      $this->config,
      $this->taxType->id()
    );
    $plugin_collection->set('avatax', $this->plugin);
    $this->taxType->set('pluginCollection', $plugin_collection);
    $this->assertEquals($this->taxType->getPlugin(), $this->plugin);

    $this->container->get('state')->set('avatax_plugin_collection:' . $this->taxType->id(), $plugin_collection);
  }

}
